# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
import argparse
import collections
from typing import Type

from manifesttool.mtool.asn1 import v3, v1
from manifesttool.mtool.asn1.manifest_codec import ManifestAsnCodecBase


class ManifestVersion:
    VERSIONS = collections.OrderedDict(
        [
            (v3.ManifestAsnCodecV3.get_name(), v3.ManifestAsnCodecV3),
            (v1.ManifestAsnCodecV1.get_name(), v1.ManifestAsnCodecV1)
        ]
    )

    @classmethod
    def list_names(cls):  # pragma: no cover
        return cls.VERSIONS.keys()

    @classmethod
    def list_codecs(cls):
        return cls.VERSIONS.values()

    @classmethod
    def from_string(cls, _str: str) -> \
            Type[ManifestAsnCodecBase]:  # pragma: no cover
        return cls.VERSIONS[_str]

    @classmethod
    def get_default(cls) -> Type[ManifestAsnCodecBase]:  # pragma: no cover
        return next(iter(cls.VERSIONS.values()))

# pylint: disable=too-few-public-methods
class StoreManifestVersion(argparse.Action):  # pragma: no cover
    def __call__(self, parser, namespace, values, option_string=None):
        prospective = values
        try:
            setattr(namespace, self.dest,
                    ManifestVersion.from_string(prospective))
        except KeyError as ex:
            raise argparse.ArgumentTypeError(
                'invalid manifest schema version') from ex



# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import argparse
import enum
import logging
import sys

from manifesttool import __version__
from manifesttool.mtool.actions.create import CreateAction
from manifesttool.mtool.actions.parse import ParseAction
from manifesttool.mtool.actions.public_key import PublicKeyAction
from manifesttool.mtool.actions.schema import PrintSchemaAction
from manifesttool.mtool.asn1.v1 import ManifestAsnCodecV1
from manifesttool.mtool.asn1.v3 import ManifestAsnCodecV3

logger = logging.getLogger('manifest-tool')


class Actions(enum.Enum):
    CREATE = 'create'
    CREATE_V1 = 'create-v1'
    PARSE = 'parse'
    SCHEMA = 'schema'
    PUB_KEY = 'public-key'


def get_parser():
    parser = argparse.ArgumentParser(
        description='Tool for creating, signing and verifying '
                    'manifest files for running Pelion Device '
                    'management update campaigns.',
        add_help=False
    )

    parser.add_argument(
        '-h',
        '--help',
        action='help',
        help='Show this help message and exit.'
    )
    parser.add_argument(
        '--version',
        action='version',
        version='Manifest-Tool version {}'.format(__version__),
        help='Show program\'s version number and exit.'
    )

    parser.add_argument(
        '-q', '--quiet',
        action='store_true',
        help='Print error logs only.'
    )

    parser.add_argument(
        '--debug',
        action='store_true',
        help='Print exception info upon exiting.'
    )

    actions_parser = parser.add_subparsers(title='Commands', dest='action')
    actions_parser.required = True

    create_parser = actions_parser.add_parser(
        Actions.CREATE.value,
        help='Create a manifest.',
        description='Create a manifest.',
        add_help=False
    )
    CreateAction.register_parser_args(
        create_parser, ManifestAsnCodecV3.get_name())

    create_parser = actions_parser.add_parser(
        Actions.CREATE_V1.value,
        help='Create a V1 schema manifest.',
        description='Create a V1 schema manifest.',
        add_help=False
    )
    CreateAction.register_parser_args(
        create_parser, ManifestAsnCodecV1.get_name())

    verify_parser = actions_parser.add_parser(
        Actions.PARSE.value,
        help='Parse and verify a manifest against the input '
             'validation schema.',
        description='Parse and verify a manifest against the input '
                    'validation schema.',
        add_help=False
    )
    ParseAction.register_parser_args(verify_parser)

    schema_parser = actions_parser.add_parser(
        Actions.SCHEMA.value,
        help='Print the input validation schema.',
        description='Print the input validation schema.',
        add_help=False
    )
    PrintSchemaAction.register_parser_args(schema_parser)

    public_key_parser = actions_parser.add_parser(
        Actions.PUB_KEY.value,
        help='Create a public key file containing a key in '
             'uncompressed point format.',
        description='Create a public key file containing a key in '
                    'uncompressed point format.',
        add_help=False
    )
    PublicKeyAction.register_parser_args(public_key_parser)

    return parser


def entry_point(argv=sys.argv[1:]):  # pylint: disable=dangerous-default-value
    parser = get_parser()
    args = parser.parse_args(argv)

    logging.basicConfig(
        stream=sys.stdout,
        format='%(asctime)s %(levelname)s %(message)s',
        level=logging.ERROR if args.quiet else logging.DEBUG
    )
    try:
        action = Actions(args.action)
        if action == Actions.PARSE:
            ParseAction.entry_point(args)
        elif action == Actions.CREATE:
            CreateAction.entry_point(args, ManifestAsnCodecV3)
        elif action == Actions.CREATE_V1:
            CreateAction.entry_point(args, ManifestAsnCodecV1)
        elif action == Actions.SCHEMA:
            PrintSchemaAction.entry_point(args)
        elif action == Actions.PUB_KEY:
            PublicKeyAction.entry_point(args)
        else:
            # will never get here
            raise AssertionError('Invalid action')
    except Exception as ex:  # pylint: disable=broad-except
        logger.error(
            str(ex),
            exc_info=args.debug
        )
        return 1
    return 0


if __name__ == '__main__':
    raise SystemExit(entry_point())



# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
import argparse
import logging
import time
from pathlib import Path
from typing import Type

import yaml

from manifesttool.dev_tool import defaults
from manifesttool.common.common_helpers import get_argument_path
from manifesttool.common.common_helpers import get_non_negative_int_argument
from manifesttool.mtool.actions import semantic_version_arg_factory
from manifesttool.mtool.actions import semver_as_tuple_arg_factory
from manifesttool.mtool.actions.create import CreateAction
from manifesttool.mtool.asn1 import ManifestAsnCodecBase
from manifesttool.mtool.payload_format import PayloadFormat

logger = logging.getLogger('manifest-dev-tool-create')


def register_parser(parser: argparse.ArgumentParser,
                    schema_version: str,
                    is_update_parser: bool = False):

    required = parser.add_argument_group('required arguments')
    optional = parser.add_argument_group('optional arguments')

    if not is_update_parser:
        required.add_argument(
            '-u', '--payload-url',
            help='Address from which the device downloads '
                 'the candidate payload.',
            required=True
        )

    required.add_argument(
        '-p', '--payload-path',
        help='Local path to the candidate payload file.',
        type=get_argument_path,
        required=True
    )

    if not is_update_parser:
        required.add_argument(
            '-o', '--output',
            help='Output manifest filename.',
            type=argparse.FileType('wb'),
            required=True
        )

    if schema_version == 'v1':

        version_group = optional.add_mutually_exclusive_group()
        version_group.add_argument(
            '-v', '--fw-version',
            type=get_non_negative_int_argument,
            help='Version number (integer) of the candidate image. '
                 'Default: current epoch time.',
            default=int(time.time())
        )
        version_group.add_argument(
            '--fw-migrate-ver',
            type=semver_as_tuple_arg_factory,
            help='Version number of the candidate image in '
                 'SemVer format. NOTE: Use to upgrade from '
                 'v1 manifest schema to a later schema.'
        )

        optional.add_argument(
            '-r', '--priority',
            type=get_non_negative_int_argument,
            help='Set update priority >=0.'
        )
    else:  # v3
        if is_update_parser:
            optional.add_argument(
                '-e', '--encrypt-payload',
                action='store_true',
                help='Encrypt the payload the device downloads.',
            )
        else:
            optional.add_argument(
                '--encrypted-digest',
                type=str,
                help='Encrypted payload digest (32-byte HEX string).'
                     ' Use only if the candidate payload is encrypted.',
            )
            optional.add_argument(
                '--encrypted-size',
                type=get_non_negative_int_argument,
                help='Encrypted payload size.'
                     ' Use only if the candidate payload is encrypted.',
            )
        optional.add_argument(
            '-v', '--fw-version',
            type=semantic_version_arg_factory,
            help='Version number of the candidate image in SemVer format. '
                 'Min. 0.0.1, max 999.999.999. '
                 'Must be bigger than the version currently '
                 'in the device(s).',
        )
        optional.add_argument(
            '--component-name',
            default='MAIN',
            metavar='NAME',
            help='The name of the component to be udpated. '
                 'Must correspond to an existing component '
                 'name on target devices. [Default: MAIN].'
        )
        if is_update_parser:
            optional.add_argument(
                '-u', '--use-short-url',
                action='store_true',
                help='Use a short candidate payload URL in the manifest. '
                     'Note: the device must be configured to use CoAP.'
            )
        optional.add_argument(
            '-m', '--sign-image',
            action='store_true',
            help='Sign image. Use only when the bootloader on a device '
                 'expects a signed FW image.'
        )
        optional.add_argument(
            '-r', '--priority',
            type=get_non_negative_int_argument,
            help='Set update priority >=0. [Default: 0].',
            default=0
        )
        optional.add_argument(
            '--combined-image',
            action='store_true',
            help='Use combined package prepared by manifest-package-tool'
        )

    optional.add_argument(
        '-d', '--vendor-data',
        help='Path to a vendor custom data file - to be passed to '
             'the target devices.',
        type=get_argument_path
    )

    optional.add_argument(
        '--cache-dir',
        help='Tool\'s cache directory. '
             'Must match the directory used by the "init" command.',
        type=Path,
        default=defaults.BASE_PATH
    )

    optional.add_argument(
        '-h',
        '--help',
        action='help',
        help='Show this help message and exit.'
    )

def set_payload_format(
    delta_meta_file,
    combined_package,
    encrypted_digest
):

    if delta_meta_file.is_file():
        if encrypted_digest:
            raise AssertionError('Unexpected combination (Delta+Encrypted)')
        payload_format = PayloadFormat.PATCH
    elif combined_package is True:
        if encrypted_digest:
            payload_format = PayloadFormat.ENCRYPTED_COMBINED
        else:
            payload_format = PayloadFormat.COMBINED
    else:
        if encrypted_digest:
            payload_format = PayloadFormat.ENCRYPTED_RAW
        else:
            payload_format = PayloadFormat.RAW

    return payload_format

def create_dev_manifest(
        dev_cfg: dict,
        manifest_version: Type[ManifestAsnCodecBase],
        vendor_data_path: Path,
        payload_path: Path,
        payload_url: str,
        encrypted_digest: str,
        encrypted_size: int,
        priority: int,
        fw_version: str,
        sign_image: bool,
        component: str,
        combined_package: bool
):
    key_file = Path(dev_cfg['key_file'])
    if not key_file.is_file():
        raise AssertionError('{} not found'.format(key_file.as_posix()))
    key_data = key_file.read_bytes()

    payload_file = payload_path
    delta_meta_file = payload_file.with_suffix('.yaml')

    payload_format = \
        set_payload_format(delta_meta_file, combined_package, encrypted_digest)

    input_cfg = {
        'vendor': {
            'vendor-id': dev_cfg['vendor-id']
        },
        'device': {
            'class-id': dev_cfg['class-id']
        },
        'payload': {
            'url': payload_url,
            'file-path': payload_path.as_posix(),
            'format': payload_format.value
        }
    }

    if encrypted_digest:
        input_cfg['payload']['encrypted'] = {
            'digest': encrypted_digest,
            'size': encrypted_size
        }

    if priority is not None:
        input_cfg['priority'] = priority

    if manifest_version.get_name() != 'v1':
        input_cfg['sign-image'] = sign_image
        input_cfg['component'] = component

    if vendor_data_path:
        input_cfg['vendor']['custom-data-path'] = vendor_data_path.as_posix()

    # logger.debug('input_cfg=\n%s', yaml.dump(input_cfg))

    manifest_bin = CreateAction.do_create(
        pem_key_data=key_data,
        input_cfg=input_cfg,
        fw_version=fw_version,
        update_certificate=Path(dev_cfg['certificate']),
        asn1_codec_class=manifest_version
    )
    logger.info(
        'Created manifest in %s schema for %s update campaign',
        manifest_version.get_name(),
        'delta' if payload_format == PayloadFormat.PATCH else 'full'
    )
    return manifest_bin

def bump_version(sem_ver: str):
    nibbles = sem_ver.split('.')
    for i in range(2, -1, -1):
        nibble = int(nibbles[i]) + 1
        if nibble <= 999:
            nibbles[i] = nibble
            break
        nibbles[i] = 0
    return semantic_version_arg_factory('{}.{}.{}'.format(*nibbles))

def load_cfg_and_get_fw_ver(
        args,
        manifest_version: Type[ManifestAsnCodecBase]
):
    if not args.cache_dir.is_dir():
        raise AssertionError('Tool cache directory is missing. '
                             'Execute "init" command to create it.')

    component_name = getattr(args, 'component_name', 'MAIN')
    fw_migrate_ver = getattr(args, 'fw_migrate_ver', None)
    cache_fw_version_file = args.cache_dir / defaults.UPDATE_VERSION
    cached_versions = {}
    fw_sem_ver = None

    # load dev_cfg
    with (args.cache_dir / defaults.DEV_CFG).open('rt') as fh:
        dev_cfg = yaml.safe_load(fh)

    # load cached_versions
    if 'v1' not in manifest_version.get_name() or fw_migrate_ver:
        if cache_fw_version_file.is_file():
            with cache_fw_version_file.open('rt') as fh:
                cached_versions = yaml.safe_load(fh)

    # set fw_version
    if 'v1' not in manifest_version.get_name():
        fw_version = args.fw_version if args.fw_version else \
            bump_version(cached_versions.get(component_name, '0.0.1'))
        fw_sem_ver = fw_version
        logger.info('FW version: %s', fw_version)
    elif fw_migrate_ver:
        fw_version, fw_sem_ver = args.fw_migrate_ver
        logger.info('FW version: %d (%s)', fw_version, fw_sem_ver)
    else:
        fw_version = args.fw_version
        logger.info('FW version: %d', fw_version)

    # save cached_versions
    if fw_sem_ver:
        cached_versions[component_name] = fw_sem_ver
        with cache_fw_version_file.open('wt') as fh:
            yaml.dump(cached_versions, fh)

    return dev_cfg, fw_version

def entry_point(
        args,
        parser: argparse.ArgumentParser,
        manifest_version: Type[ManifestAsnCodecBase]
):
    encrypted_digest = getattr(args, 'encrypted_digest', None)
    encrypted_size = getattr(args, 'encrypted_size', None)

    if encrypted_digest or encrypted_size:
        if not encrypted_digest or not encrypted_size:
            parser.error(
                'require --encrypted-digest and --encrypted-size '
                'or none of those.'
            )

    dev_cfg, fw_version = \
        load_cfg_and_get_fw_ver(args, manifest_version)

    manifest_bin = create_dev_manifest(
        dev_cfg=dev_cfg,
        manifest_version=manifest_version,
        vendor_data_path=args.vendor_data,
        payload_path=args.payload_path,
        payload_url=args.payload_url,
        encrypted_digest=encrypted_digest,
        encrypted_size=encrypted_size,
        priority=args.priority,
        fw_version=fw_version,
        sign_image=getattr(args, 'sign_image', False),
        component=getattr(args, 'component_name', None),
        combined_package=getattr(args, 'combined_image', False)
    )

    with args.output as fh:
        fh.write(manifest_bin)



# ----------------------------------------------------------------------------
# Copyright 2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
import argparse
import logging
import sys
import enum

from manifesttool.package_tool.actions.create import CreateAction
from manifesttool.package_tool.actions.parse import ParseAction
from manifesttool import __version__

logger = logging.getLogger("manifest-package-tool")

class Actions(enum.Enum):
    CREATE = 'create'
    PARSE = 'parse'

def get_parser():
    parser = argparse.ArgumentParser(
        description='Tool for creating and verifying combined package files '
                    'used for full update campaigns.',
        add_help=False
    )
    required = parser.add_argument_group('required arguments')

    required.add_argument(
        '-h', '--help',
        action='help',
        help='Show this help message and exit.'
    )

    actions_parser = parser.add_subparsers(title='Commands', dest='action')
    actions_parser.required = True

    create_parser = actions_parser.add_parser(
        Actions.CREATE.value,
        help='Create a package.',
        description='Create a package.',
        add_help=False
    )

    CreateAction.register_parser_args(
        create_parser)

    verify_parser = actions_parser.add_parser(
        Actions.PARSE.value,
        help='Parse and verify a package against the input '
             'validation schema.',
        description='Parse and verify a package against the input '
                    'validation schema.',
        add_help=False
    )
    ParseAction.register_parser_args(verify_parser)
    return parser


def entry_point(argv=sys.argv[1:]):  # pylint: disable=dangerous-default-value
    parser = get_parser()
    args = parser.parse_args(argv)

    logging.basicConfig(
        stream=sys.stdout,
        format='%(asctime)s %(levelname)s %(message)s',
        level=logging.DEBUG
    )
    try:
        action = Actions(args.action)
        if action == Actions.PARSE:
            ParseAction.entry_point(args)
        elif action == Actions.CREATE:
            CreateAction.entry_point(args)
        else:
            # will never get here
            raise AssertionError('Invalid action')
    except Exception as ex:  # pylint: disable=broad-except
        logger.error(
            str(ex),
            # exc_info=args.debug
        )
        return 1
    return 0


if __name__ == '__main__':
    raise SystemExit(entry_point())



# ----------------------------------------------------------------------------
# Copyright 2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
import argparse
from pathlib import Path

def get_argument_path(value):
    """
        Construct Path to an existing file for an argument
        :param value: input string
        :return: constructed Path with semantics
        appropriate for the operating systems
    """
    arg_path = Path(value)
    if not arg_path.is_file():
        raise argparse.ArgumentTypeError(
            'File "{}" is not found'.format(value)
        )
    return arg_path

def get_file_size(path_file: Path):
    return Path(path_file).stat().st_size

def get_non_negative_int_argument(value: str) -> int:
    """
    Construct non negative integer value for an argument
    :param value: input string
    :return: integer value
    """
    int_value = None
    try:
        int_value = int(value)
    except ValueError:
        pass
    if int_value is None or int_value < 0:
        raise argparse.ArgumentTypeError(
            '"{}" is an invalid non-negative integer value'.format(value))
    return int_value



# ----------------------------------------------------------------------------
# Copyright 2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import argparse
import logging
from pathlib import Path
import jsonschema
import yaml
from manifesttool.common.common_helpers import get_argument_path
from manifesttool.package_tool.asn1.package_encoder import DescriptorAsnCodec
from manifesttool.package_tool.package_format.tar_package \
    import PackageFormatTar

PTOOL_PATH = Path(__file__).resolve().parent.parent
logger = logging.getLogger('manifest-package-tool-create')

def _get_alignment_int_argument(value: str) -> int:
    """
    Construct non negative integer value for an argument
    :param value: input string
    :return: integer value
    """
    int_value = None
    try:
        int_value = int(value)
    except ValueError:
        pass
    if int_value is None or int_value < 1:
        raise argparse.ArgumentTypeError(
            '"{}" is an invalid alignment value, should be >0'.format(value))
    return int_value

class CreateAction:

    @staticmethod
    def register_parser_args(
            parser: argparse.ArgumentParser):
        required = parser.add_argument_group('required arguments')
        optional = parser.add_argument_group('optional arguments')

        required.add_argument(
            '-c',
            '--config',
            help='Path to the package tool configuration file.',
            metavar='YAML',
            type=get_argument_path,
            required=True
        )

        required.add_argument(
            '-f',
            '--format',
            help='Package format type.',
            choices=['tar'],  # Add bin when implemented
            default='tar'
        )

        required.add_argument(
            '-a',
            '--image-alignment-size',
            help='Candidate storage read size, \
                used for image alignment. Relevant for embedded devices.',
            default=1,
            type=_get_alignment_int_argument
        )

        required.add_argument(
            '-o',
            '--output',
            help='Output package filename.',
            type=Path,
            required=True
        )

        optional.add_argument(
            '-h',
            '--help',
            action='help',
            help='Show this help message and exit.'
        )

    @staticmethod
    def do_create(
        input_cfg: dict,
        output_file,
        format_type
    ):

        # validate input against input-schema.json
        schema_path = PTOOL_PATH / 'package-schema.json'
        with schema_path.open('rb') as fh:
            input_schema = yaml.safe_load(fh)
            jsonschema.validate(input_cfg, input_schema)

        # create asn1 descriptor
        codec = DescriptorAsnCodec()
        asn1der = codec.encode_package_descriptor(input_cfg)

        if 'tar' in format_type:
            package_format = PackageFormatTar()
        else:
            logger.error("Package tool supports only tar format")
            raise NotImplementedError

        package_format.create_package(output_file, input_cfg, asn1der)

    @classmethod
    def entry_point(
            cls,
            args: argparse.Namespace
    ) -> None:

        output_file = args.output

        with open(args.config, "rb") as config_fh:
            input_cfg = yaml.safe_load(config_fh)

        cls.do_create(input_cfg, output_file, args.format)



# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
# Auto-generated by asn1ate v.0.6.0 from manifest_v3.asn
# (last modified on 2021-06-22 12:17:55.466130)

from pyasn1.type import univ, char, namedtype, namedval, tag, constraint, useful


class PayloadMetadata(univ.Sequence):
    pass


PayloadMetadata.componentType = namedtype.NamedTypes(
    namedtype.NamedType('installed-size', univ.Integer()),
    namedtype.NamedType('installed-digest', univ.OctetString()),
    namedtype.OptionalNamedType('precursor-digest', univ.OctetString())
)


class Manifest(univ.Sequence):
    pass


Manifest.componentType = namedtype.NamedTypes(
    namedtype.NamedType('vendor-id', univ.OctetString()),
    namedtype.NamedType('class-id', univ.OctetString()),
    namedtype.NamedType('update-priority', univ.Integer()),
    namedtype.NamedType('component-name', char.UTF8String()),
    namedtype.NamedType('payload-version', char.UTF8String()),
    namedtype.NamedType('payload-digest', univ.OctetString()),
    namedtype.NamedType('payload-size', univ.Integer()),
    namedtype.NamedType('payload-uri', char.UTF8String()),
    namedtype.NamedType('payload-format', univ.Enumerated(namedValues=namedval.NamedValues(('raw-binary', 1), ('arm-patch-stream', 5), ('combined', 6), ('encrypted-raw', 257), ('encrypted-combined', 262)))),
    namedtype.NamedType('installed-signature', univ.OctetString()),
    namedtype.OptionalNamedType('payload-metadata', PayloadMetadata()),
    namedtype.OptionalNamedType('vendor-data', univ.OctetString())
)


class SignedResource(univ.Sequence):
    pass


SignedResource.componentType = namedtype.NamedTypes(
    namedtype.NamedType('manifest-version', univ.Enumerated(namedValues=namedval.NamedValues(('v3', 3)))),
    namedtype.NamedType('manifest', Manifest()),
    namedtype.NamedType('signature', univ.OctetString())
)





# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
import argparse
import base64
import hashlib
import logging
import re
import sys
from mmap import mmap, ACCESS_READ
from pathlib import Path

import yaml
from manifesttool.common.common_helpers import get_argument_path
from manifesttool import __version__
from manifesttool import armbsdiff

logger = logging.getLogger("manifest-delta-tool")

def _block_size_factory(value):
    prospective = None
    try:
        prospective = int(value)
    except ValueError:
        pass
    if not prospective or prospective < 128:
        raise argparse.ArgumentTypeError(
            "{} is invalid - must be at least 128".format(value))
    return prospective


def digest_file(file_path: Path):
    read_block_size = 65536
    with file_path.open('rb') as fh:
        hash_ctx = hashlib.sha256()
        buf = fh.read(read_block_size)
        while buf:
            hash_ctx.update(buf)
            buf = fh.read(read_block_size)
        file_len = fh.tell()
    return base64.b64encode(hash_ctx.digest()), file_len


def size_check(new_size, delta_size, threshold):
    """
    Assert new fw image is smaller than delta file

    :param new_size: full new FW image size
    :param delta_size: delta file size
    :param threshold: size difference threshold for aborting the generation.
                      in case evaluates to False - size check will be aborted.
    """

    diff = 100 * float(delta_size) / float(new_size)

    if threshold and diff >= threshold:
        raise AssertionError(
            'Difference with delta image and update image is more than '
            '{} percent! Percentage is: {:.2f}'.format(threshold, diff))


def get_version_string_from_bin_file(fname: Path):
    with fname.open('rb') as fh:
        with mmap(fh.fileno(), 0, access=ACCESS_READ) as my_mmap:
            match = re.search(b'(PELION/BSDIFF\\d{3})', my_mmap)
            if not match:
                raise AssertionError("Version details not found")
            return match.group(0).decode('utf-8')


def check_bsdiff_bspatch_versions(original_image_path: Path):
    bsdiff_version = armbsdiff.get_version()
    logger.info("Current tool version %s", bsdiff_version)
    bspatch_version = get_version_string_from_bin_file(original_image_path)

    if bsdiff_version != bspatch_version:
        logger.error(
            'Bspatch version in {} is incomatible with this version '
            'of delta-tool.')

        logger.error(
            'Original image version is: %s',
            bspatch_version.decode("utf-8").split('/')[1]
        )

        logger.error(
            'Current bsdiff version is: %s',
            bsdiff_version.decode("utf-8").split('/')[1]
        )

        raise AssertionError('bsdiff/bspatch version mismatch')


def generate_delta(
        orig_fw: Path,
        new_fw: Path,
        output_delta_file: Path,
        block_size,
        threshold
):
    check_bsdiff_bspatch_versions(orig_fw)

    original_digest, _ = digest_file(orig_fw)
    new_digest, new_size = digest_file(new_fw)

    if original_digest == new_digest:
        logger.warning('New and old file are binary same. This will generate '
                       'delta that will not change the original image. '
                       'This is probably a mistake')

    try:
        armbsdiff.generate(
            orig_fw.as_posix(),
            new_fw.as_posix(),
            output_delta_file.as_posix(),
            block_size
        )

        _, delta_size = digest_file(output_delta_file)

        size_check(new_size, delta_size, threshold)
    except AssertionError:
        if output_delta_file.is_file():
            output_delta_file.unlink()
        raise

    delta_cfg_file = output_delta_file.with_suffix('.yaml')
    config = {
        'installed-digest': new_digest,
        'installed-size': new_size,
        'precursor-digest': original_digest
    }
    with delta_cfg_file.open('wt') as fh:
        yaml.dump(config, fh)


def get_parser():
    parser = argparse.ArgumentParser(
        description='Generate delta patch files to be used for '
                    'delta update campaigns.',
        add_help=False
    )
    required = parser.add_argument_group('required arguments')
    optional = parser.add_argument_group('optional arguments')

    required.add_argument(
        '-c', '--current-fw',
        type=get_argument_path,
        help='Path to the currently installed firmware image, without '
             'headers, for delta update calculation.',
        required=True
    )
    required.add_argument(
        '-n', '--new-fw',
        type=get_argument_path,
        help='Path to the candidate image.',
        required=True
    )

    required.add_argument(
        '-o', '--output',
        type=Path,
        help='Output delta patch filename. '
             'NOTE: The delta tool generates an additional '
             'configuration file with the same name but with a \'.yaml\' '
             'extension. The manifest tool needs both files, '
             'but only this output file must '
             'be uploaded to Pelion storage.',
        required=True
    )

    optional.add_argument(
        '-b', '--block-size',
        type=_block_size_factory,
        help='Compression block size algorithm. '
             'A greater size provides better '
             'compression, but consumes more memory on the device. '
             'Default is 512 bytes. Minimum is 128 bytes. '
             'NOTE: This value MUST be aligned with the '
             'network (COAP/HTTP) buffer size used for download.',
        default=512
    )

    size_group = optional.add_mutually_exclusive_group()

    # must be first in a group for default value to be set properly
    size_group.add_argument(
        '-t', '--threshold',
        type=int,
        choices=range(30, 100),
        metavar="[30-100]",
        default=60,
        help='The ratio of the delta patch size compared to the '
             'candidate image size above which to raise an exception. '
             'Default is 60.'
    )
    size_group.add_argument(
        '--skip-size-check',
        action='store_false',
        dest='threshold',
        help='Skip threshold validations.'
    )

    optional.add_argument(
        '--debug',
        action='store_true',
        help='Print exception info upon exiting.'
    )

    optional.add_argument(
        '-h',
        '--help',
        action='help',
        help='Show this help message and exit.'
    )
    optional.add_argument(
        '--version',
        action='version',
        version='Manifest-Tool version {}'.format(__version__),
        help='Show program\'s version number and exit.'
    )

    return parser


def entry_point(argv=sys.argv[1:]):  # pylint: disable=dangerous-default-value
    parser = get_parser()
    args = parser.parse_args(argv)

    logging.basicConfig(
        stream=sys.stdout,
        format='%(asctime)s %(levelname)s %(message)s',
        level=logging.DEBUG
    )

    try:
        generate_delta(
            orig_fw=args.current_fw,
            new_fw=args.new_fw,
            output_delta_file=args.output,
            block_size=args.block_size,
            threshold=args.threshold
        )
    except Exception as ex:  # pylint: disable=broad-except
        logger.error(
            str(ex),
            exc_info=args.debug
        )
        return 1
    return 0


if __name__ == '__main__':
    raise SystemExit(entry_point())



# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
from pathlib import Path

MTOOL_PATH = Path(__file__).resolve().parent.parent


class PrintSchemaAction:
    @staticmethod
    def print_schema():
        schema_file = MTOOL_PATH / 'manifest-input-schema.json'
        print(schema_file.read_text())

    @staticmethod
    def register_parser_args(parser):
        optional = parser.add_argument_group('optional arguments')

        optional.add_argument(
            '-h',
            '--help',
            action='help',
            help='Show this help message and exit.'
        )

    @classmethod
    def entry_point(cls, _):
        cls.print_schema()



# ----------------------------------------------------------------------------
# Copyright 2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
import tarfile
import logging
from manifesttool.common.common_helpers import get_argument_path
from manifesttool.package_tool.package_format.tar_package \
    import PackageFormatTar

class ParseAction:

    logger = logging.getLogger('manifest-package-parse')

    @staticmethod
    def register_parser_args(parser):
        required = parser.add_argument_group('required arguments')
        optional = parser.add_argument_group('optional arguments')

        required.add_argument(
            '-p', '--package',
            help='Path to the package file.',
            type=get_argument_path,
            required=True
        )

        optional.add_argument(
            '-h',
            '--help',
            action='help',
            help='Show this help message and exit.'
        )

    # pylint: disable=too-many-branches
    @staticmethod
    def do_parse(package):

        if tarfile.is_tarfile(package):
            package_format = PackageFormatTar()
        else:
            logging.error("Package tool supports only tar format")
            raise NotImplementedError

        package_format.parse_package(package)

    @classmethod
    def entry_point(cls, args):

        logging.info(args.package)
        cls.do_parse(args.package)



# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
from pathlib import Path

import yaml
BASE_PATH = Path('.manifest-dev-tool')
UPDATE_PUBLIC_KEY_CERT = 'dev.cert.der'
UPDATE_PRIVATE_KEY = 'dev.key.pem'
DEV_CFG = 'dev.cfg.yaml'
UPDATE_RESOURCE_C = 'update_default_resources.c'
CLOUD_CFG = 'dev.cloud_cfg.yaml'
DEV_README = 'README.txt'
API_GW = 'https://api.us-east-1.mbedcloud.com'
UPDATE_VERSION = 'update.version.yaml'

PELION_GW_PATH = Path.home() / '.pelion-dev-presets.yaml'
PELION_GW = None
if PELION_GW_PATH.is_file():
    with PELION_GW_PATH.open('rb') as fh:
        PELION_GW = yaml.safe_load(fh)



# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
import argparse
from pathlib import Path

from manifesttool.mtool import ecdsa_helper
from manifesttool.common.common_helpers import get_argument_path

class PublicKeyAction:
    @staticmethod
    def register_parser_args(parser: argparse.ArgumentParser):
        required = parser.add_argument_group('required arguments')
        optional = parser.add_argument_group('optional arguments')

        required.add_argument(
            'private_key',
            help='Path to a private key PEM file.',
            type=get_argument_path,
        )

        required.add_argument(
            '-o', '--out',
            help='Output public key filename.',
            type=Path,
            required=True
        )

        optional.add_argument(
            '-h',
            '--help',
            action='help',
            help='Show this help message and exit.'
        )

    @classmethod
    def get_key(cls, private_key_bytes: bytes):
        public_key = ecdsa_helper.public_key_from_private(private_key_bytes)
        public_key_bytes = ecdsa_helper.public_key_to_bytes(public_key)
        return public_key_bytes

    @classmethod
    def entry_point(cls, args):
        private_key_bytes = cls.get_key(args.private_key.read_bytes())
        args.out.write_bytes(private_key_bytes)



# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

import argparse
import logging
import time
from pathlib import Path
from typing import Type

import jsonschema
import yaml

from manifesttool.common.common_helpers import get_argument_path
from manifesttool.common.common_helpers import get_non_negative_int_argument
from manifesttool.mtool.actions import semantic_version_arg_factory
from manifesttool.mtool.actions import semver_as_tuple_arg_factory
from manifesttool.mtool.asn1 import ManifestAsnCodecBase
from manifesttool.mtool.asn1.v1 import ManifestAsnCodecV1
from manifesttool.mtool import ecdsa_helper

MTOOL_PATH = Path(__file__).resolve().parent.parent

logger = logging.getLogger('manifest-tool-create')


class CreateAction:

    @staticmethod
    def register_parser_args(
            parser: argparse.ArgumentParser, schema_version: str):
        required = parser.add_argument_group('required arguments')
        optional = parser.add_argument_group('optional arguments')

        required.add_argument(
            '-c',
            '--config',
            help='Path to the manifest tool configuration file.',
            metavar='YAML',
            type=argparse.FileType('rb'),
            required=True
        )

        required.add_argument(
            '-k',
            '--key',
            help='Path to the PEM format private key file.',
            metavar='KEY',
            type=argparse.FileType('rb'),
            required=True
        )

        if schema_version == 'v1':
            version_group = optional.add_mutually_exclusive_group()
            version_group.add_argument(
                '-v', '--fw-version',
                type=get_non_negative_int_argument,
                help='Version number (integer) of the candidate image. '
                     'Default: current epoch time.',
                default=int(time.time())
            )
            version_group.add_argument(
                '--fw-migrate-ver',
                type=semver_as_tuple_arg_factory,
                help='Version number of the candidate image in '
                     'SemVer format. NOTE: Use to upgrade from '
                     'v1 manifest schema to a later schema.'
            )

            required.add_argument(
                '--update-certificate',
                type=get_argument_path,
                help='Path to the update certificate file.',
                required=True
            )
        else:
            required.add_argument(
                '-v', '--fw-version',
                type=semantic_version_arg_factory,
                help='Version number of the candidate image in SemVer format. '
                     'Min. 0.0.1, max 999.999.999. '
                     'Must be bigger than the version currently '
                     'in the device(s).',
                required=True
            )

        required.add_argument(
            '-o',
            '--output',
            help='Output manifest filename.',
            type=argparse.FileType('wb'),
            required=True
        )

        optional.add_argument(
            '-h',
            '--help',
            action='help',
            help='Show this help message and exit.'
        )

    @staticmethod
    def do_create(
            pem_key_data: bytes,
            input_cfg: dict,
            fw_version,
            update_certificate: Path,
            asn1_codec_class: Type[ManifestAsnCodecBase]
    ) -> bytes:
        assert fw_version is not None

        codec = asn1_codec_class()

        # validate input against manifest-input-schema.json
        schema_path = MTOOL_PATH / 'manifest-input-schema.json'
        with schema_path.open('rb') as fh:
            input_schema = yaml.safe_load(fh)
            if isinstance(codec, ManifestAsnCodecV1):
                # priority field is optional for v1
                # delete it from required list
                input_schema['required'].remove('priority')
            jsonschema.validate(input_cfg, input_schema)

        raw_signature = True
        if isinstance(codec, ManifestAsnCodecV1):
            raw_signature = False
            cert_data = update_certificate.read_bytes()
            codec.set_update_certificate(cert_data)
        installed_digest = codec.process_input_config(fw_version, input_cfg)

        if input_cfg.get('sign-image'):
            if isinstance(codec, ManifestAsnCodecV1):
                raise AssertionError(
                    'sign-image is unexpected for manifest schema v1')
            signature = ecdsa_helper.ecdsa_sign_prehashed(
                installed_digest, pem_key_data)
            codec.set_image_signature(
                ecdsa_helper.signature_der_to_raw(signature)
            )
        else:
            codec.set_image_signature(bytes())

        der_manifest = codec.get_signed_data()

        signature = ecdsa_helper.ecdsa_sign(der_manifest, pem_key_data)
        if raw_signature:
            signature = ecdsa_helper.signature_der_to_raw(signature)
        manifest_bin = codec.get_der_signed_resource(signature)

        logger.info('Attention: When updating Mbed OS devices,'
                    ' candidate features must match the device\'s'
                    ' bootloader features. Incompatibility may'
                    ' result in damaged devices.')

        return manifest_bin

    @classmethod
    def entry_point(
            cls,
            args: argparse.Namespace,
            asn1_codec: Type[ManifestAsnCodecBase]
    ) -> None:

        input_cfg = yaml.safe_load(args.config)

        if getattr(args, 'fw_migrate_ver', None):
            fw_version = args.fw_migrate_ver[0]
        else:
            fw_version = args.fw_version

        manifest_bin = cls.do_create(
            pem_key_data=args.key.read(),
            input_cfg=input_cfg,
            fw_version=fw_version,
            update_certificate=getattr(args, 'update_certificate', None),
            asn1_codec_class=asn1_codec
        )
        with args.output as fh:
            fh.write(manifest_bin)



# ----------------------------------------------------------------------------
# Copyright 2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
import abc
import logging
from typing import TypeVar

DESCRIPTOR_FILE_NAME = "_desc_"

PackageFormatBaseType = TypeVar(
    'PackageFormatBaseType', bound='PackageFormatBase')

logger = logging.getLogger('package-format')

class PackageFormatBase(abc.ABC):

    @abc.abstractmethod
    def parse_package(self, package_file):
        raise NotImplementedError

    @abc.abstractmethod
    def create_package(self, output_file, input_cfg, asn1der):
        raise NotImplementedError



# ----------------------------------------------------------------------------
# Copyright 2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
import logging
import tempfile
from pathlib import Path
import tarfile
from manifesttool.package_tool.package_format.package_format \
    import PackageFormatBase
from manifesttool.package_tool.package_format.package_format \
    import DESCRIPTOR_FILE_NAME
from manifesttool.package_tool.asn1.package_encoder \
    import DescriptorAsnCodec

TAR_TOOL_PATH = Path(__file__).resolve().parent.parent
DESC_FILE = TAR_TOOL_PATH / DESCRIPTOR_FILE_NAME

class PackageFormatTar(PackageFormatBase):
    logger = logging.getLogger('manifest-tar-package')

    def create_package(self, output_file, input_cfg: dict, asn1der):

        with tempfile.NamedTemporaryFile(delete=False) as fh:
            fh.write(asn1der)
            fh.close()

            with tarfile.open(output_file, "w:") as tar_handle:

                # Add descriptor file to the package
                tar_handle.add(fh.name, DESCRIPTOR_FILE_NAME)
                Path(fh.name).unlink()

                logging.debug('add files')
                # Add all images to the package
                for image in input_cfg['images']:
                    tar_handle.add(image['file_name'], image['sub_comp_name'])

    @classmethod
    def parse_package(cls, package_file):

        with tarfile.open(package_file, "r:") as tar_arch:
            logging.info("Contents of the tar package - ")
            for tarinfo in tar_arch:
                logging.info('File name : %s', tarinfo.name)
                if tarinfo.name in DESCRIPTOR_FILE_NAME:
                    with tar_arch.extractfile(DESCRIPTOR_FILE_NAME) as desc_fh:
                        asn1der = desc_fh.read()
                    asn1_dict = DescriptorAsnCodec.decode(asn1der)

        logging.info("Information of update images:")
        for img in asn1_dict['descriptors-array']:
            logging.info(img)



# ----------------------------------------------------------------------------
# Copyright 2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
import logging
from collections import OrderedDict
from pyasn1.codec.native.encoder import encode as native_encoder
from pyasn1.codec.der import decoder as der_decoder
from pyasn1.codec.der import encoder as der_encoder
from pyasn1.type.univ import SequenceOf
from manifesttool.package_tool.asn1 import package_schema
from manifesttool.common.common_helpers import get_file_size

class DescriptorAsnCodec:

    logger = logging.getLogger('descriptor-codec')

    def __init__(self, dom=None):

        if dom:
            self.dom = dom
        else:
            self.dom = package_schema.Descriptor()
            self.dom['descriptors-array'] = SequenceOf()

    def set_asn1_descriptor(self, input_cfg: dict):

        self.dom['num-of-images'] = len(input_cfg['images'])

        for image in input_cfg['images']:
            img_descriptor = self.get_img_descriptor(image)
            self.dom['descriptors-array'].append(img_descriptor)

    @staticmethod
    def get_img_descriptor(image):

        image_descriptor = package_schema.ImgDescriptor()
        image_descriptor['id'] = image['sub_comp_name']
        image_descriptor['vendor-data'] = image['vendor_data']
        image_descriptor['vendor-data-size'] = len(image['vendor_data'])
        image_descriptor['image-size'] = get_file_size(image['file_name'])

        return image_descriptor

    def encode_package_descriptor(self, input_cfg: dict) -> bytes:

        self.set_asn1_descriptor(input_cfg)
        return der_encoder.encode(self.dom)

    @classmethod
    def decode(
            cls,
            data: bytes) -> OrderedDict:

        package_dom = \
            der_decoder.decode(data, asn1Spec=package_schema.Descriptor())[0]

        return native_encoder(package_dom)



# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
import hashlib
import logging
import time
from collections import OrderedDict
from typing import Optional

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives.asymmetric import ec
from pyasn1.codec.der import decoder as der_decoder
from pyasn1.codec.der import encoder as der_encoder
from pyasn1.codec.native.encoder import encode as native_encoder
from pyasn1.type.univ import SequenceOf

from manifesttool.mtool import ecdsa_helper
from manifesttool.mtool.asn1.manifest_codec import ManifestAsnCodecBase
from manifesttool.mtool.asn1.v1 import manifest_schema_v1 as schema


class ManifestAsnCodecV1(ManifestAsnCodecBase):
    VERSION = 'v1'

    logger = logging.getLogger('v1-manifest-codec')

    def __init__(self, dom=None):

        if dom:
            self.dom = dom
        else:
            self.dom = schema.Resource()
            self.dom['resourceType'] = 'manifest'
            self.dom['resource']['manifest'] = schema.Manifest()
            self.dom['resource']['manifest']['manifestVersion'] = 'v1'
            self.dom['resource']['manifest']['nonce'] = b''
            self.dom['resource']['manifest']['deviceId'] = b''
            self.dom['resource']['manifest']['vendorInfo'] = b''
            self.dom['resource']['manifest']['applyImmediately'] = True
            self.dom['resource']['manifest']['encryptionMode']['enum'] = \
                'none-ecc-secp256r1-sha256'
            self.dom['resource']['manifest'][
                'payload'] = schema.PayloadDescription()
            self.dom['resource']['manifest']['payload'][
                'storageIdentifier'] = "default"
            self.dom['resource']['manifest']['payload'][
                'reference'] = schema.ResourceReference()
            self.fingerprint = None

    @classmethod
    def get_name(cls) -> str:
        return cls.VERSION

    def get_der_signed_resource(self, signature: bytes) -> bytes:
        hash_ctx = hashlib.sha256()
        hash_ctx.update(self.get_signed_data())
        signed_digest = hash_ctx.digest()

        resource_signature_dom = schema.ResourceSignature()
        resource_signature_dom['hash'] = signed_digest
        signature_block = schema.SignatureBlock()
        signature_block['signature'] = signature
        signature_block['certificates'] = SequenceOf()
        certificate_reference = schema.CertificateReference()
        certificate_reference['fingerprint'] = self.fingerprint
        certificate_reference['uri'] = ''
        signature_block['certificates'].append(certificate_reference)
        resource_signature_dom['signatures'] = SequenceOf()
        resource_signature_dom['signatures'].append(signature_block)

        signed_resource_dom = schema.SignedResource()
        signed_resource_dom['resource'] = self.dom
        signed_resource_dom['signature'] = resource_signature_dom

        return der_encoder.encode(signed_resource_dom)

    def get_signed_data(self) -> bytes:
        return der_encoder.encode(self.dom)

    @classmethod
    def decode(
            cls,
            data: bytes,
            verification_key: Optional[ec.EllipticCurvePublicKey]
    ) -> OrderedDict:
        signed_resource_dom = der_decoder.decode(
            data,
            asn1Spec=schema.SignedResource())[0]
        if not verification_key:
            return native_encoder(signed_resource_dom)

        codec = ManifestAsnCodecV1(dom=signed_resource_dom['resource'])
        signed_data = codec.get_signed_data()
        signature = bytes(
            signed_resource_dom['signature']['signatures'][0]['signature'])

        try:
            ecdsa_helper.ecdsa_verify(
                public_key=verification_key,
                signed_data=signed_data,
                signature=signature
            )
            cls.logger.info('Manifest Signature verified!')
        except InvalidSignature as ex:
            raise AssertionError(
                'Manifest Signature verification failed') from ex
        return native_encoder(signed_resource_dom)

    def set_payload_version(self, version: str):
        if version:
            try:
                version_number = int(version)
            except ValueError as ex:
                raise AssertionError(
                    'invalid version {} - version must be '
                    'a valid positive integer'.format(version)
                ) from ex
        else:
            version_number = int(time.time())
        self.dom['resource']['manifest']['timestamp'] = version_number

    def set_update_priority(self, priority: int):
        self.dom['resource']['manifest']['priority'] = priority

    def set_vendor_id(self, vendor_id: bytes):
        self.dom['resource']['manifest']['vendorId'] = vendor_id

    def set_vendor_data(self, data: bytes):
        self.dom['resource']['manifest']['vendorInfo'] = data

    def set_class_id(self, class_id: bytes):
        self.dom['resource']['manifest']['classId'] = class_id

    def set_device_id(self, device_id: bytes):
        self.dom['resource']['manifest']['classId'] = device_id

    def set_payload_fingerprint(self, digest: bytes, size: int):
        self.dom['resource']['manifest']['payload']['reference'][
            'hash'] = digest
        self.dom['resource']['manifest']['payload']['reference']['size'] = size

    def set_payload_uri(self, uri: str):
        self.dom['resource']['manifest']['payload']['reference']['uri'] = uri

    def set_payload_format(self, payload_format: str):
        _format = payload_format
        if payload_format == 'arm-patch-stream':
            _format = 'bsdiff-stream'
        self.dom['resource']['manifest']['payload']['format']['enum'] = _format

    def set_payload_metadata(
            self,
            installed_digest: bytes,
            installed_size: int,
            precursor_digest: bytes
    ):
        self.dom['resource']['manifest']['payload'][
            'installedDigest'] = installed_digest
        self.dom['resource']['manifest']['payload'][
            'installedSize'] = installed_size
        self.dom['resource']['manifest']['precursorDigest'] = precursor_digest

    def set_update_certificate(self, cert_data: bytes):
        hash_ctx = hashlib.sha256()
        hash_ctx.update(cert_data)
        self.fingerprint = hash_ctx.digest()

    def set_image_signature(self, signature: bytes):
        pass  # do nothing here

    def set_component_name(self, component: str):
        pass



# ----------------------------------------------------------------------------
# Copyright 2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
# Auto-generated by asn1ate v.0.6.0 from package.asn
# (last modified on 2021-06-24 10:25:24.529484)

from pyasn1.type import univ, char, namedtype, constraint


class Bytes(univ.OctetString):
    pass


class Int(univ.Integer):
    pass


class String(char.UTF8String):
    pass


class ImgDescriptor(univ.Sequence):
    pass


ImgDescriptor.componentType = namedtype.NamedTypes(
    namedtype.NamedType('id', String()),
    namedtype.NamedType('vendor-data', Bytes()),
    namedtype.NamedType('vendor-data-size', Int()),
    namedtype.NamedType('image-size', Int())
)


class Descriptor(univ.Sequence):
    pass


Descriptor.componentType = namedtype.NamedTypes(
    namedtype.NamedType('num-of-images', \
        univ.Integer().subtype(subtypeSpec=constraint.ValueRangeConstraint(0, 255))),
    namedtype.NamedType('descriptors-array', univ.SequenceOf(componentType=ImgDescriptor()))
)



# ----------------------------------------------------------------------------
# Copyright 2019-2021 Pelion
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric import utils as ec_utils
from cryptography.utils import int_to_bytes

def _load_ec_private_key(pem_key_data) -> ec.EllipticCurvePrivateKey:
    private_key = serialization.load_pem_private_key(
        pem_key_data,
        password=None,
        backend=default_backend()
    )
    if isinstance(private_key, ec.EllipticCurvePrivateKey):
        return private_key
    raise TypeError()

def ecdsa_sign_prehashed(digest, pem_key_data) -> bytes:
    # pylint: disable=no-value-for-parameter
    private_key: ec.EllipticCurvePrivateKey
    private_key = _load_ec_private_key(pem_key_data)
    return private_key.sign(
        digest, ec.ECDSA(ec_utils.Prehashed(hashes.SHA256()))
    )


def ecdsa_sign(data_to_sign, pem_key_data) -> bytes:
    # pylint: disable=no-value-for-parameter
    private_key = _load_ec_private_key(pem_key_data)
    return private_key.sign(
        data_to_sign, ec.ECDSA(hashes.SHA256())
    )

def public_key_from_private(pem_priv_key_data) -> ec.EllipticCurvePublicKey:
    private_key = serialization.load_pem_private_key(
        pem_priv_key_data,
        password=None,
        backend=default_backend()
    )
    return private_key.public_key()

def public_key_from_certificate(certificate_data) -> ec.EllipticCurvePublicKey:
    cert = x509.load_der_x509_certificate(
        certificate_data, default_backend())
    return cert.public_key()


def public_key_from_bytes(public_key_data) -> ec.EllipticCurvePublicKey:
    return ec.EllipticCurvePublicKey.from_encoded_point(
        curve=ec.SECP256R1(),
        data=public_key_data
    )

def public_key_to_bytes(public_key) -> bytes:
    return public_key.public_bytes(
        encoding=serialization.Encoding.X962,
        format=serialization.PublicFormat.UncompressedPoint
    )


def signature_raw_to_der(raw_signature):
    return ec_utils.encode_dss_signature(
        r=int.from_bytes(raw_signature[:32], byteorder='big'),
        s=int.from_bytes(raw_signature[32:], byteorder='big')
    )


def signature_der_to_raw(signature):
    # pylint: disable=invalid-name
    r, s = ec_utils.decode_dss_signature(signature)
    return int_to_bytes(r, 32) + int_to_bytes(s, 32)


def ecdsa_verify_prehashed(public_key, digest, signature):
    public_key.verify(
        signature,
        digest,
        ec.ECDSA(ec_utils.Prehashed(hashes.SHA256()))
    )


def ecdsa_verify(public_key, signed_data, signature):
    public_key.verify(
        signature,
        signed_data,
        ec.ECDSA(hashes.SHA256())
    )
